#!/usr/bin/env python3

import sys
import argparse

import json


def device_name(type_id):
    device_types = {
        # gateways
        71: "Gateway CANopen",
        72: "Gateway CC-Link",
        73: "Gateway DeviceNet",
        74: "Gateway EtherCAT",
        75: "Gateway Ethernet/IP",
        76: "Gateway Powerlink",
        77: "Gateway Profibus",
        78: "Gateway Profinet RT",
        79: "Gateway Profinet IRT",
        80: "Gateway CANopen Master",
        81: "Gateway SercosIII",
        82: "Gateway Serial",
        85: "Gateway EtherCAT Master",
        92: "Gateway Modbus RTU",
        93: "Gateway Modbus TCP",
        100: "Gateway DMX",
        # revpi modules (pibridge)
        96: "RevPi DIO",
        97: "RevPi DI",
        98: "RevPi DO",
        103: "RevPi AIO",
        118: "RevPi MIO",
        137: "RevPi RO",
    }

    return device_types.get(type_id, "Unknown Device")


class FWUFile:
    SIZE_KBMAGIC_FIELD = 2
    SIZE_LENGTH_FIELD = 4
    SIZE_FPGA_HEADER = 8
    SIZE_FW_HEADER = 22

    def __init__(self, fwu_file: str):
        self.filename = fwu_file

        if self._read(0, 2) != b"KB":
            raise Exception("FWU file is missing magic number")

        header_length = self._read_int(2, 4)

        self.fw_header_offset = self.SIZE_KBMAGIC_FIELD + self.SIZE_LENGTH_FIELD
        self.data_offset = self.fw_header_offset + header_length

        # fwu version 3 contains the ST and FPGA firmware
        if self.fwu_revision == 3:
            fpga_length_offset = self.data_offset - self.SIZE_FPGA_HEADER
            fpga_length = self._read_int(fpga_length_offset, 4)
            self.data_offset += fpga_length

    def _read(self, offset: int, length: int = 1) -> bytes:
        with open(self.filename, "rb") as fh:
            fh.seek(offset)

            return fh.read(length)

    def _read_int(self, offset: int, length: int = 1) -> int:
        value = self._read(offset, length)

        return int.from_bytes(value, "little")

    @property
    def module_type(self) -> int:
        return self._read_int(self.fw_header_offset, 2)

    @property
    def hardware_revision(self) -> int:
        return self._read_int(self.fw_header_offset + 2, 2)

    @property
    def fwu_revision(self) -> int:
        return self._read_int(self.fw_header_offset + 4, 2)

    @property
    def version_major(self) -> int:
        return self._read_int(self.data_offset + 29)

    @property
    def version_minor(self) -> int:
        return self._read_int(self.data_offset + 30)

    @property
    def version(self) -> str:
        return f"{self.version_major}.{self.version_minor}"

    @property
    def svn_tag(self) -> int:
        if self.git_hash:
            # if a git hash is present, than there is no svn tag
            return ""

        return self._read_int(self.data_offset + 32)

    @property
    def git_hash(self) -> int:
        git = self._read_int(self.data_offset + 32, 4)

        if git >> 16 == 0:
            # git hash must be 32 bits
            return ""

        return hex(git)[2:].zfill(8)

    def to_json(self) -> str:
        data = {
            "fwu_revision": self.fwu_revision,
            "module_type": self.module_type,
            "module_name": device_name(self.module_type),
            "version": self.version,
            "version_major": self.version_major,
            "version_minor": self.version_minor,
            "hardware_revision": self.hardware_revision,
            "svn_tag": self.svn_tag,
            "git_hash": self.git_hash,
        }

        return json.dumps(data)


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("fwu_file")

    args = parser.parse_args()

    try:
        fwu = FWUFile(args.fwu_file)
        print(fwu.to_json())
    except Exception as e:
        print(str(e), file=sys.stderr)

        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())
